!*************************************************************************
      subroutine BPCGAP(hnew,ibound,cr,cc,cv,hcof,rhs,
     1      kkstp,kkper,icnvg,mxiter,ncol,nrow,
     2      nlay,nodes,in,iout)
!
!*************************************************************************
!         |        The PRECGN solver was created       |                 *
!         |   by D. Braess, Ruhr-Universitaet Bochum   |                 *
!         |                  Sept. 1993                |                 *
!*************************************************************************
!
!*************************************************************************
!         |     MODFLOW PRECGN solver setup module     |                 *
!         |                  (BPCGAP)                  |                 *
!         |        was created by Jon Paul Jones       |                 *
!         |           University of Waterloo           |                 *
!         |           version 2.0 Jan., 2000           |                 *
!*************************************************************************
!
!*************************************************************************
!     BPCGAP converts MODFLOW data into a form that can be sent to the   *
!     PRECGN solver. PRECGN solves an equation of the form               *
!     [A]{X}={B} where [A] is a rank 2 array and both {B} and {X} are    *
!     rank 1 arrays. Square brackets [ ] denote a rank two array and     *
!     curly brackets { } denote a rank 1 array in these notes. [A]       *
!     consists of the head coefficients z,b,d,e,f,h, and s. For example, *
!     a 4 col. by 4 row by 2 layer a matrix would look like:             *
!                                                                        *
!                                                                        *
!                     **********************************                 *
!                     *ef  h           s               *                 *
!                     *def  h           s              *                 *
!                     * def  h           s             *                 *
!                     *  def  h           s            *                 *
!                     *b  def  h           s           *                 *
!                     * b  def  h           s          *                 *
!                     *  b  def  h           s         *                 *
!                     *   b  def  h           s        *                 *
!                     *    b  def  h           s       *                 *
!                     *     b  def  h           s      *                 *
!                     *      b  def  h           s     *                 *
!                     *       b  def  h           s    *                 *
!                     *        b  def  h           s   *                 *
!                     *         b  def  h           s  *                 *
!                     *          b  def  h           s *                 *
!                     *           b  def  h           s*                 *
!                     *z           b  def  h           *                 *
!                     * z           b  def  h          *                 *
!                     *  z           b  def  h         *                 *
!                     *   z           b  def  h        *                 *
!                     *    z           b  def  h       *                 *
!                     *     z           b  def  h      *                 *
!                     *      z           b  def  h     *                 *
!                     *       z           b  def  h    *                 *
!                     *        z           b  def  h   *                 *
!                     *         z           b  def  h  *                 *
!                     *          z           b  def  h *                 *
!                     *           z           b  def  h*                 *
!                     *            z           b  def  *                 *
!                     *             z           b  def *                 *
!                     *              z           b  def*                 *
!                     *               z           b  de*                 *
!                     **********************************                 *
!                                                                        *
!                                                                        *
!     As explained in the SIP solver documentation (in MODFLOW manual)   *
!     [A] is symmetric about its main diagonal e and so only f, h, and   *
!     s (which correspond to elements of the {CR},{CC}, and {CV} arrays  *
!     respectively) and the main diagonal e will be passed to the PRECGN *
!     solver. Like MODFLOW, the PRECGN solver only utilizes rank 1       *
!     arrays and so [A] must be condensed to a rank 1 array before being *
!     passed on to the solver. Also, [A] is partitioned to remove the    *
!     type 1 (constant head) and inactive cells before it is decomposed  *
!     to a rank 1 array. It should be noted that the PRECGN solver       *
!     expects the main diagonal e to be non-zero and positive.           *
!       In the PRECGN solver the rank 2 head coefficient array [A]       *
!     is condensed into a rank 1 array also named {A}. The PRECGN solver *
!     utilizes skyline numbering. This means {A} stores values by        *
!     starting with the value at e and then recording all of the off-    *
!     diagonal values directly above e (ie. off-diagonals lying within   *
!     the same column above e.) The row numbers of e and the off-        *
!     diagonals are stored in the array {iaa} and the row numbers of     *
!     the e values are stored in the array {ind}. After storing and      *
!     recording all the data with respect to a particular e, The skyline *
!     numbering then continues storing values and row numbers with the   *
!     next e.                                                            *
!       The length of the MODFLOW array {X} is unaffected by this        *
!     module because all of the new arrays generated in BPCGAP are       *
!     allocated when needed and then deallocated when no longer needed,  *
!     Thereby freeing up computer memory used by the arrays. The only    *
!     alteration to the array {X} is in the values of hnew and rhs.      *
!*************************************************************************
!
      parameter (dzero=0.d0)
      real(8) :: hnew,bhnew,fhnew,zhnew,shnew,hhnew,dhnew,e,f,h,s,b,d,z,
     1      bb,dd,zz,rrhs,diff_h,biggest_h,hclose
      integer :: length_of_aold,laa,icnvg,iout
      integer(4), dimension(:), allocatable ::ind,iaa,kblock,iaa_old
      dimension :: cr(nodes),cc(nodes),ibound(nodes),cv(nodes),
     1      hcof(nodes),rhs(nodes),hnew(nodes)
      real(8), dimension(:), allocatable :: hnew_condensed,a,aa,a_old,
     1      rhs_condensed
!
!*****Make some preliminary definitions and allocations.
      nrc=nrow*ncol ! equal to the number of nodes in one layer
      length_of_aold=nodes+(nodes-1)+(nodes-ncol)+(nodes-nrc)
      allocate (a_old(1:length_of_aold))
      allocate (iaa_old(1:length_of_aold))
!
!*************************************************************************
!                                                                        *
!           If the cell in question is surrounded by inactive cells      *
!                 then make the cell inactive.                           *
!                                                                        *
!*************************************************************************
!
      do 5000 k=1,nlay      ! nlay = number of layers
      do 5000 i=1,nrow      ! nrow = number of rows
      do 5000 j=1,ncol      ! ncol = number of columns
!
      n=j+(i-1)*ncol+(k-1)*nrc      ! the unpartitioned matrix index
      if (ibound(n).eq.0) go to 5000
!
!*****Calculate 1 dimensional subscripts for locating the 6
!*****surrounding cells.
      nrn=n+ncol  ! cell index of adjacent cell in next row
      nrl=n-ncol  ! cell index of adjacent cell in last row
      ncn=n+1     ! cell index of adjacent cell in next column
      ncl=n-1     ! cell index of adjacent cell in last column
      nln=n+nrc   ! cell index of adjacent cell in next layer
      nll=n-nrc   ! cell index of adjacent cell in last layer
!
!*****Calculate 1 dimensional subscripts for conductance to the 6
!*****surrounding cells.
      ncf=n       ! conductance index of adjacent cell in next column 
      ncd=n-1     ! conductance index of adjacent cell in last column
      nrb=n-ncol  ! conductance index of adjacent cell in last row
      nrh=n       ! conductance index of adjacent cell in next row
      nls=n       ! conductance index of adjacent cell in next layer
      nlz=n-nrc   ! conductance index of adjacent cell in last layer
!
      b=dzero
      if (i.eq.1.or.ibound(nrl).eq.0) go to 5010
      b=cc(nrb)   ! conductance of adjacent cell in last row
5010  continue
!
      h=dzero
      if (i.eq.nrow.or.ibound(nrn).eq.0) go to 5020
      h=cc(nrh)   ! conductance of adjacent cell in next row
5020  continue
!
      d=dzero      
      if (j.eq.1.or.ibound(ncl).eq.0) go to 5030
      d=cr(ncd)   ! conductance of adjacent cell in last column
5030  continue
!
      f=dzero
      if (j.eq.ncol.or.ibound(ncn).eq.0) go to 5040
      f=cr(ncf)   ! conductance of adjacent cell in next column
5040  continue
!
      z=dzero
      if (k.eq.1.or.ibound(nll).eq.0) go to 5050
      z=cv(nlz)   ! conductance of adjacent cell in last layer
5050  continue
!
      s=dzero
      if (k.eq.nlay.or.ibound(nln).eq.0) go to 5060
      s=cv(nls)   ! conductance of adjacent cell in next layer
5060  continue
!*****Make cell inactive if all surrounding cells are inactive.
      if(b+h+d+f+z+s.eq.0.) then
          ibound(n)=0
          hcof(n)=0.
          rhs(n)=0.
      endif
!
5000  continue
!*************************************************************************
!
!*************************************************************************
!         Find the number of unconstrained head nodes.                   *
!*************************************************************************
      n1=0
      do 5070 m1=1,nodes
      if(ibound(m1).gt.0) n1=n1+1
5070  continue
!
!*************************************************************************
!       Allocate arrays that need prior knowledge of the number of       *
!                       unconstrained nodes.                             *
!*************************************************************************
      allocate (hnew_condensed(1:n1))
      allocate (rhs_condensed(1:n1))
      allocate (ind(1:(n1+1)))
      n2=0
      kk=0
!
!*************************************************************************
!                         |   The Main Loop   |                          *
!*************************************************************************
!                          Loop's function:                              *
!                 -generate the condensed rhs array                      *
!                 -generate the condensed hnew array                     *
!                 -determine e, f, h, s, d, z, and b for unpartitioned   *
!                  nodes                                                 *
!                 -fill the dummy arrays for the solver                  *
!*************************************************************************
!
      do 5080 k=1,nlay  ! nlay = number of layers
      do 5080 i=1,nrow  ! nrow = number of rows
      do 5080 j=1,ncol  ! ncol = number of columns
!
      n=j+(i-1)*ncol+(k-1)*nrc    ! the unpartitioned matrix index
!
      if (ibound(n).le.0) go to 5080  ! skip calculating partitioned nodes
!
!*****Calculate 1 dimensional subscripts for locating the 6
!*****surrounding cells.
      nrn=n+ncol  ! cell index of adjacent cell in next row
      nrl=n-ncol  ! cell index of adjacent cell in last row
      ncn=n+1     ! cell index of adjacent cell in next column
      ncl=n-1     ! cell index of adjacent cell in last column
      nln=n+nrc   ! cell index of adjacent cell in next layer
      nll=n-nrc   ! cell index of adjacent cell in last layer
!
!*****Calculate 1 dimensional subscripts for conductance to the 6
!*****surrounding cells.
      ncf=n       ! conductance index of adjacent cell in next column 
      ncd=n-1     ! conductance index of adjacent cell in last column
      nrb=n-ncol  ! conductance index of adjacent cell in last row
      nrh=n       ! conductance index of adjacent cell in next row
      nls=n       ! conductance index of adjacent cell in next layer
      nlz=n-nrc   ! conductance index of adjacent cell in last layer
!
!*****Evaluate the main diagonal and rhs.
      bb=dzero
      b=dzero
      bhnew=dzero      
      if (i.eq.1.or.ibound(nrl).le.0) go to 5090
      bb=cc(nrb)      ! conductance of adjacent cell in last row      
5090  if (i.eq.1.or.ibound(nrl).eq.0) go to 5100
      b=cc(nrb)      ! conductance of adjacent cell in last row
      if (ibound(nrl).lt.0) bhnew=b*hnew(nrl)
5100  continue
!
      h=dzero
      hhnew=dzero
      if (i.eq.nrow.or.ibound(nrn).eq.0) go to 5110
      h=cc(nrh)   ! conductance of adjacent cell in next row
      if (ibound(nrn).lt.0) hhnew=h*hnew(nrn)
5110  continue
!
      dd=dzero
      d=dzero      
      dhnew=dzero      
      if (j.eq.1.or.ibound(ncl).le.0) go to 5120
      dd=cr(ncd)  ! conductance of adjacent cell in last column
5120  if (j.eq.1.or.ibound(ncl).eq.0) go to 5130
      d=cr(ncd)   ! conductance of adjacent cell in last column
      if (ibound(ncl).lt.0) dhnew=d*hnew(ncl)
5130  continue
!
      f=dzero
      fhnew=dzero
      if (j.eq.ncol.or.ibound(ncn).eq.0) go to 5140
      f=cr(ncf)   ! conductance of adjacent cell in next column
      if (ibound(ncn).lt.0) fhnew=f*hnew(ncn)
5140  continue
!
      zz=dzero
      z=dzero
      zhnew=dzero      
      if (k.eq.1.or.ibound(nll).le.0) go to 5150
      zz=cv(nlz)  ! conductance of adjacent cell in last layer      
5150  if (k.eq.1.or.ibound(nll).eq.0) go to 5160
      z=cv(nlz)   ! conductance of adjacent cell in last layer
      if (ibound(nll).lt.0) zhnew=z*hnew(nll)
5160  continue
!
      s=dzero
      shnew=dzero
      if (k.eq.nlay.or.ibound(nln).eq.0) go to 5170
      s=cv(nls)   ! conductance of adjacent cell in next layer
      if (ibound(nln).lt.0) shnew=s*hnew(nln)
5170  continue
!
      e=-z-b-d-f-h-s+hcof(n)
      rrhs=rhs(n)
      rhs(n)=rrhs-fhnew-hhnew-shnew-bhnew-dhnew-zhnew
!
!*************************************************************************
!
!*************************************************************************
!     -generate condensed rhs array (condensed by partitioning)          *
!     -generate condensed hnew array (condensed by partitioning)         *
!************************************************************************* 
      n2=n2+1
      rhs_condensed(n2)=-rhs(n)
      hnew_condensed(n2)=hnew(n)
!
!*****Begin filling dummy solver arrays.
      kk=kk+1
      a_old(kk)=-e      ! storage of main diagonal into [a]
      ind(n2)=kk        ! storage of main diagonal positions in [a]
      iaa_old(kk)=n2    ! storage of row number of each element of [a]
!                  
!*****Storage of f into {a}.
!*************************************************************************
!     Note: in this skyline numbering scheme, it is more efficient to    * 
!     use the term to the left of the main diagonal instead of the term  *
!     directly above the main diagonal. So at this point dd will be      *
!     substituted for the f term.                                        *
!*************************************************************************
      if(n2.eq.1) go to 5080
      kk=kk+1
      a_old(kk)=-dd
      iaa_old(kk)=(n2-1)
!
!*****Storage of h into {a}.
!*************************************************************************
!     Note: in this skyline numbering scheme, it is more efficient to    * 
!     use the term to the left of the main diagonal instead of the term  *
!     directly above the main diagonal. So at this point bb will be      *
!     substituted for the h term.                                        *
!*************************************************************************
      nhh=0
      if(n.lt.(ncol+1)) go to 5080
      do 5180 nh=(n-ncol),n
        if(ibound(nh).le.0)nhh=nhh+1
5180  continue
      kk=kk+1
      a_old(kk)=-bb
      iaa_old(kk)=(n2-ncol+nhh)
!
!*****Storage of s into {a}.
!*************************************************************************
!     Note: in this skyline numbering scheme, it is more efficient to    * 
!     use the term to the left of the main diagonal instead of the term  *
!     directly above the main diagonal. So at this point zz will be      *
!     substituted for the s term.                                        *
!*************************************************************************
      nss=0
      if(n.lt.(nrc+1)) go to 5080
      do 5190 ns=(n-nrc),n
        if(ibound(ns).le.0)nss=nss+1
5190  continue
      kk=kk+1
      a_old(kk)=-zz
      iaa_old(kk)=(n2-nrc+nss)
!
5080  continue
!
!*************************************************************************
!             Prepare the data to be sent to the PRECGN solver.          *
!*************************************************************************
      lblo=n2
      ind(n2+1)=kk+1
      new=1
      if(((3*n2)+kk).lt.(6*n2))laa=6*n2
      if(((3*n2)+kk).ge.(6*n2))laa=(3*n2)+kk
      allocate (kblock(1:lblo))
      allocate (a(1:kk))
      allocate (aa(1:laa))
      allocate (iaa(1:kk))
      a=dzero
      aa=dzero
      iaa=0
!
!*****Fill the solver arrays with the dummy array values.
      do 5200 n=1,kk
      a(n)=a_old(n)
      iaa(n)=iaa_old(n)
5200  continue
      deallocate (iaa_old,a_old)  ! free up memory
!
!*************************************************************************
!     The following parameters are specified by the user & read in:      *
!       a) itmax: this is the maximum number of iterations               *
!       b) hclose: this is the convergence criteria for the head values   *
!                                                                        *
!      Suggested default values include:                                 *
!          itmax=8000                                                    *
!          hclose=1.e-8                                                   *
!*************************************************************************
      if (kkstp.eq.1.and.kkper.eq.1) then
!*****The following line will read in the above values (free format).
         read (in,*) itmax, hclose
!
!*************************************************************************
!*****           Print a message identifying BPCG package            *****
!*************************************************************************
         write(iout,1)
 1    format(//1x,'PRECGN -- conjugate gradient solution package using 
     1matrix solver written by D. Braess',/11x,
     2'conjugate gradient preconditioning based on block-wise ssor-relax
     3ation ...'/11x,'MODFLOW-solver interface (BPCGAP) version 2.0 
     4written by J. Jones, January 2000')
!*****Write the input data to the main MODFLOW output file.
           write(iout,3) itmax,hclose
 3    format(//1x,'PRECGN input parameters:'/
     1     7x,' maximum number of iterations (itmax) = ',i5/
     2     7x,' convergence criteria (hclose) = ',1pe8.1)
      end if
!
!*************************************************************************
!          All information needed by the solver has been generated       *
!                    and so the solver will be called.                   *
!*************************************************************************
!
      write (iout,5210)
5210  format (//'PRECGN solver has been called.'/)
	icnvg=1
      call PRECGN(a,n1,iaa,ind,rhs_condensed,hnew_condensed,aa,laa,
     1      kblock,lblo,new,iout,itmax,hclose,icnvg)
      write (iout,5220)
5220  format (/'program back from PRECGN solver'/)
!
      deallocate (a,ind,iaa,aa,kblock)      ! free up memory
!*****Update the rhs and hnew arrays.
      n=0
      do 5230 nn=1,nodes
      if (ibound(nn).gt.0) then
         n=n+1
         rhs(nn)=-rhs_condensed(n)
         hnew(nn)=hnew_condensed(n)
      endif
5230  continue
      deallocate (rhs_condensed,hnew_condensed)      ! free up memory
      return
      end
!
!*************************************************************************
!****************|    PRECGN solver code starts here    |*****************
!*************************************************************************
!
!*************************************************************************
      subroutine PRECGN(a,n,iaa,ind,b,x,aa,laa,kblock,lblo,new,iout,
     1    itmax,hclose,icnvg)
!*************************************************************************
!     Matrix solver for [A]*{X}={B} with conjugate gradient              *
!     preconditioning through incomplete lu-decomposition using          *
!     block-wise ssor-relaxation tridiagonal blocks will be determined   *
!*************************************************************************
!*************************************************************************
!                                            D.Braess      Sept. 1993    *
!   fast solver for [A]*{X} = {B}            Ruhr-Universitaet Bochum    *
!                                                                        *
!   {a}:      expected in condensed form, will not be destroyed          *
!   n:        the number of rows in [a]                                  *
!   {iaa}:    contains the row number of each element of [a]             *
!   {ind}:    contains positions of main diagonal elements of [a]        *
!   {b}:      the flux vector on the right hand side                     *
!   {x}:      the head vector                                            *
!   {aa}:     additional storage array of length laa.ge. 6*n             *
!   laa:      = 3*nodes + the length of [a]                              *
!   {kblock}: additional storage array of length lblo                    *
!   lblo:     reccomended that lblo = n                                  *
!   new:      = 1:  new matrix, defined start value for x                *
!   iout:     unit number of the MODFLOW output file                     *
!   itmax:    maximum number of iterations                               *
!   hclose:    head convergence criteria                                  *
!                                                                        *
!   sharing of additional storage array: {aa(kmax)},{g(n)},{h(n)},{d(n)} *
!*************************************************************************
!*************************************************************************
!                    program cgicc - notes                               *
!              ===============================                           *
!                                                                        *
!       warnings                                                         *
!       --------                                                         *
!                                                                        *
!   1) pcg used ... steps                                                *
!      for a relative accuracy of only ...                               *
!                                                                        *
!      --  if this warning appears exclusively, preconditioning is not   *
!          sufficient to reach the desired accuracy using the defined    *
!          number of (iteration) steps.                                  *
!          remedy: increase the maximum number of iteration steps        *
!          ('itmax') or choose a different solver.                       *
!                                                                        *
!   2) preconditioning precision is already good                         *
!                                                                        *
!      --  the accuracy is reduced by the factor 'hclose/100',            *
!          process is stopped, to improve the iterative procedure        *
!          and to avoid truncation errors.                               *
!                                                                        *
!   3) values from beta to ...                                           *
!          indicate instable preconditioning.                            *
!                                                                        *
!      --  during the computation of conjugated directions small         *
!          denominators arise.  preconditioning is barely sufficient.    *
!                                                                        *
!   4) note: during ilu-decomposition, a reduction                       *
!          of diagonal elements by the factor ... is required.           *
!                                                                        *
!      --  due to the (incomplete) choleski-decomposition, diagonal      *
!          elements get reduced. a factor of reduction close to 1        *
!          indicates a high number of conditioning processes.            *
!          for stabilization, reduction is dampened.  this might lead to *
!          a higher number of iteration steps.                           *
!                                                                        *
!      errors                                                            *
!      ------                                                            *
!   5) pcg with icc expects additional storage array of length ...       *
!      additional storage array too small!                               *
!                                                                        *
!     --  the program requires an additional storage array of length     *
!         3n + number of effective matrix elements.                      *
!                                                                        *
!   6) pcg has been assigned a matrix with ...-th diagonal element < 0!  *
!                                                                        *
!     --  the algorithm is designated for positive definite matrices.    *
!         the diagonal elements have to be positive.                     *
!                                                                        *
!   7) step ... determined the matrix as indefinite.                     *
!                                                                        *
!     --  analysis using the correction vector results in a value        *
!         for d'ad, which is not positive.                               *
!         the matrix is not positive definite or the analysis is         *
!         sensitive with respect to truncation errors.                   *
!                                                                        *
!   8) step - spacing too large!                                         *
!                                                                        *
!     --  analysis of d'ad results in small denominators, calculation    *
!         is sensitive with respect to truncation errors.                *
!                                                                        *
!   9) 50% of maximum number of steps completed - reduction not          *
!      satisfactory!                                                     *
!                                                                        *
!     --  it can't be expected, that required accuracy will be reached   *
!         using the defined number of steps.  preconditioning not        *
!         sufficient.                                                    *
!                                                                        *
!*************************************************************************
!
!*****Partitioning of working arrays: g(n), h(n), d(n), ax(n).
      implicit none
      integer(4)  n, new, ind(1),iaa(1), laa, lblo,kblock(lblo)
      integer(4)  laamin,kmax, kg,kh,kd,kax,kdia,ksup,iout,itmax,icnvg
      real(8)   a(1),aa(1), b(n),x(n),hclose
!
      kmax   = ind(n+1) - 1
      laamin = 6*n
      if (laa.lt.laamin) go to 80
      kg   = 1
      kh   = kg + n
      kd   = kh + n
      kax  = kd + n
      kdia = kax  + n
      ksup = kdia + n
      call cgssortri (a,n,iaa,ind,b,x,aa(kg),aa(kh),aa(kd),
     1aa(kax),aa(kdia),aa(ksup),kblock,lblo,new,iout,itmax,hclose,icnvg)
      go to 90
   80 write (iout,86) laamin
   86 format (' pcg with ssor expects additonal storage array of length
     1 ',     i6 / ' additional storage array too small! ')
   90 return
      end
!
!*************************************************************************
      subroutine cgssortri (a,n,iaa,ind,b,x,g,h,d,ax,
     1                  dia,sup,kblock,lblo,new,iout,itmax,hclose,icnvg)
!
!*************************************************************************
!     General cg-method with preconditioning, here especially with       *
!     block-ssor relaxation.                                             *
!*************************************************************************
!
      implicit none
      integer(4)   itmax,icnvg
      real(8)    hclose,hclose2
      integer(4)   n, iaa(1), ind(1), new, lblo,kblock(lblo),iout
      real(8)    a(1),b(n),x(n),  g(n),h(n),d(n),ax(n),dia(n),sup(n)
      real(8)    gh,ghalt,dhd,gg, gg0,gg1,gh1,sqgg, tau,beta,betamx, bb
      integer(4)   i,k,iter
!
      hclose2=hclose**2
!
!*****Starting values and control of provided matrix.
      bb    = 0.
      call mavmukn (a,n,iaa,ind, x,g)
      do  5 i=1,n
      bb   =   bb + b(i)*b(i)
      g(i) =   g(i) - b(i)
      h(i) =   g(i)
      d(i) =   0.
      if (a(ind(i)) .le. 0.) then
         write (iout,81) i
   81 format(' pcg has been assigned a matrix with',i6,
     1       '-th main diagonal entry < 0 !')
          stop
      end if
    5 continue
!
!*****If the right side is trivial then it is not calculated.
      if (bb .eq. 0.) then
         write (iout,*)
     1 ' precg recieved input where the right side = 0'
         return
      end if
      gg0 = gg
!
      call zerltri (a,n, iaa,ind, dia,sup, kblock,lblo,iout)
!
!*****Do an iteration.
      gh   =   0.
      beta =   0.
      betamx = 0.
      do 60 k=1,itmax
      iter = k
!
!*****Now doing h = c**-1 * g.
      call ssortri (a,n, iaa,ind, g,h, ax, dia,sup, kblock,lblo)
      ghalt = gh
      gh    = 0.
      gg    = 0.
      do 14 i=1,n
      gh = gh + g(i)*h(i)
      gg = gg + g(i)*g(i)
   14 continue
      if (k .eq. 1) then
          gg1  = gg
          gh1  = gh
      else
          beta   = gh / ghalt
          betamx = max (beta,betamx)
      end if
      if (k .eq. 2)  gh1  = gh
      do 16 i=1,n
   16 d(i) = - h(i) + beta * d(i)
!
!*****Now doing h = a*d.
      call mavmukn (a,n, iaa,ind, d,h)
      dhd = 0.
      do 20 i=1,n
   20 dhd = dhd + d(i)*h(i)
      if (dhd .le. 0.) then
          write (iout,26) k
   26 format (' step ', i4,' determined the matrix to be indefinite!')
          stop
      end if
      if (dhd .lt. 1.e-6*gh) then
          write (iout,*) ' pcg: step width too big!'
          go to 90
      end if
      tau = gh / dhd
      do 30 i=1,n
      x(i) = x(i) + tau * d(i)
      g(i) = g(i) + tau * h(i)
      h(i) = g(i)
   30 continue
      if (gg .le. min (.01*gg1, hclose2*bb)
     1   .or. gg .le. .0001*hclose2*bb)           go to 99
      if (gh .lt. 0.00001*hclose2 * gh1) then
         if (gh .lt. -hclose2 * gh1) then
             write (iout,*) ' pcg: preconditioning is indefinite'
         else
             write (iout,*) ' pcg - warning:',
     1   '  preconditioning precision is already good'
         end if
         go to 90
      end if
      if (2*iter.gt.itmax .and. gg.gt.0.25*gg1) then
         write (iout,*)
     1 ' pcg: after half of the maximum step numbers still no real 
     2reduction!'
         go to 90
      end if
   60 continue
!
   90 if (gg .ge. hclose2 * bb) then
         sqgg = sqrt (gg/bb)
         write (iout,87) iter, sqgg
   87 format (' pcg used ', i4,' steps'/
     1        ' for a relative accuracy of only ',1p1e11.2)
         if (betamx .gt. 50.) write (iout,88) betamx
   88 format (' values from beta to',1pe11.3 /
     1        ' indicate unstable preconditioning')
      end if
	if (iter.eq.itmax) icnvg=0
   99 continue
      write (iout,100)iter
100   format('PRECGN solver used ',i5,' iterations')
	return
      end
!
!*************************************************************************
      subroutine mavmukn (a,n, iaa,inda, v,z)
!*************************************************************************
!
!*****Multiplication of condensed matrix with vector.
      implicit none
      integer(4)  n, iaa(1),inda(1)
      real(8)   a(1),v(n),z(n)
      integer(4)  k,kugr,kogr, i,j
      real(8)   vj,zj
!
!*****Because i<j no initializing is necessary, copy n.
      kogr = inda(1) - 1
      do 30 j=1,n
      kugr = kogr + 2
      kogr = inda(j+1)- 1
!
!*****Diagonal.
      vj   = v(j)
      zj   = a(kugr-1) * vj
      do 20 k=kugr,kogr
      i = iaa(k)
      zj   = zj   + a(k) * v(i)
      z(i) = z(i) + a(k) * vj
   20 continue
      z(j) = zj
   30 continue
      return
      end
!*************************************************************************
      subroutine ssortri (a,n, iaa,ind, b,x, ax,dia,sup, kblock,lblo)
!     block-ssor with tridiagonal blocks
!*************************************************************************
!
!*************************************************************************
!     1 step of the symetric overrelaxation to get a * x = b             *
!     while blocks are treated simultaneously,                           *
!     as long as this goes over tridiagonal matrices                     *
!     b and x are not allowed to be identical                            *
!     ax is a helping array of length n -- after: dx/omega               *
!*************************************************************************
!
      implicit none
      real(8)    omega
      parameter (omega=1.3d0)
      integer(4)   n, iaa(1), kogr, jogr
      integer(4)   ind(1), mi,k,iout
      integer(4)   j,i, nschicht, jj, lblo,kblock(lblo), kkb,ib
      real(8)    a(1), b(n),x(n),ax(n), dia(n),sup(n),  s,xj
      save      kkb
!
!*****Forward iteration.
      kogr = ind(1)-1
      do 40 ib=1,kkb
      jj       = kblock(ib)
      jogr     = kblock(ib+1)-1
      nschicht = kblock(ib+1)-jj
      do 32 j=jj,jogr
      mi   = kogr + 1
      kogr = ind(j+1) - 1
      s    = b(j)
      do 30 k=mi+min(2,j-jj+1), kogr
   30 s = s - a(k)*x(iaa(k))
      ax(j) = s
   32 continue
!
!*****Forward in block.
      x(jj) = ax(jj)
      do 34 j=jj+1,jogr
      x(j)  = ax(j) - sup(j) * x(j-1)
   34 continue
!
!*****Backward in block.
      x(jogr) = x(jogr) * dia(jogr)
      do 36 j=jogr,jj+1,-1
   36 x(j-1) = x(j-1) * dia(j-1) - sup(j) * x(j)
   40 continue
!
!*****Backward iteration.
      mi = ind(n+1)
      do 50 ib=kkb,1,-1
      jj       = kblock(ib)
      jogr     = kblock(ib+1)-1
      nschicht = kblock(ib+1)-jj
!
!*****Forward in block.
      x(jj)   = ax(jj)
      do 44 j=jj+1,jogr
   44 x(j)    = ax(j) - sup(j) * x(j-1)
!
!*****Backward in block.
      x(jogr) = x(jogr) * dia(jogr)
      do 46 j=jogr,jj+1,-1
   46 x(j-1) = x(j-1) * dia(j-1) - sup(j) * x(j)
      do 48 j=jogr,jj,-1
      kogr = mi - 1
      mi   = ind(j)
      xj   = x(j)
      do 48 k=mi+min(2,j-jj+1), kogr
      i = iaa(k)
      ax(i) = ax(i) - a(k)*xj
   48 continue
   50 continue
      return
!*************************************************************************
      entry zerltri (a,n, iaa,ind, dia,sup, kblock,lblo,iout)
!     Cholesky decomposition with diagonal blocks
!*************************************************************************
!
!*****Splitting up into a tridiagonal block.
      jj  = 0
      kkb = 0
      do 14 j=1,n
      mi = ind(j)
      if (ind(j+1).eq.mi+1 .or. iaa(mi+1).ne.j-1 .or.
     1   (ind(j+1).gt.mi+2 .and.iaa(mi+2).ge.jj)) then
!
!*****Newer block.
         kkb = kkb + 1
         jj  = j
         kblock(kkb) = j
         if (kkb .ge. lblo) then
            write (iout,*)
     1    ' cgtriadapt: dimension for number block is too small!'
            stop
         end if
      end if
      dia(j) = a(mi)
      sup(j) = a(mi+1)
   14 continue
      kblock(kkb+1) = n+1
      do 24 ib=1,kkb
      jj      = kblock(ib)
      jogr    = kblock(ib+1)-1
      dia(jj) = 1. / dia(jj)
      do 24 j=jj+1, jogr
      s      = sup(j)
      sup(j) = sup(j) * dia(j-1)
      dia(j) = 1. / (dia(j) - sup(j)*s)
   24 continue
!
!*****Overrelaxtion.
      do 26 j=1,n
   26 dia(j) = omega * dia(j)
      return
      end
